from typing import Union, List, Dict
from .plisio_enums import CryptoCurrency, FiatCurrency, OperationStatus, OperationType, PlanName

from .plisio_exceptions import (
    PlisioError,
    RequestNotProcessed,
    RequestAlreadyProcessed,
    UnknownPlisioAPIError,
    BadRequestError,
    UnauthorizedError,
    ForbiddenError,
    NotFoundError,
    MethodNotAllowedError,
    NotAcceptableError,
    UnsupportedMediaTypeError,
    UnprocessableEntityTypeError,
    RateLimitReachedError,
    InternalServerError,
    ServiceUnavailableError,
)

from .plisio_models import (
    PlisioModel,
    Balance,
    Currency,
    Invoice,
    Commission,
    Withdraw,
    Fee,
    Plan,
    FeePlan,
    Operation,
    Operations,
)

from .plisio_client import PlisioClient, PlisioAioClient

RType = Union[List['RType'], Dict[str, 'RType']]

ModelType = Union[
    'plisio.PlisioModel',
    'plisio.Balance',
    'plisio.Currency',
    'plisio.Invoice',
    'plisio.Commission',
    'plisio.Withdraw',
    'plisio.Fee',
    'plisio.Plan',
    'plisio.FeePlan',
    'plisio.Operation',
    'plisio.Operations',
]



from typing import Optional


class PlisioError(Exception):
    reason = 'An error has occurred when contacting to Plisio API'

    def __init__(self, message: Optional[str] = None):
        if message is None:
            message = self.reason
        super().__init__(message)


class RequestNotProcessed(PlisioError):
    reason = 'The request to Plisio API has not yet been processed'


class RequestAlreadyProcessed(PlisioError):
    reason = 'The request to Plisio API has already been processed'


class UnknownPlisioAPIError(PlisioError):
    pass


class BadRequestError(PlisioError):
    """
    400 Bad Request.
    INVALID_REQUEST.
    Request is not well-formed, syntactically incorrect, or violates schema.
    """
    reason = 'Request is not well-formed, syntactically incorrect, or violates schema.'


class UnauthorizedError(PlisioError):
    """
    401 Unauthorized.
    AUTHENTICATION_FAILURE.
    Authentication failed due to invalid authentication credentials.
    """
    reason = 'Authentication failed due to invalid authentication credentials.'


class ForbiddenError(PlisioError):
    """
    403 Forbidden.
    NOT_AUTHORIZED.
    Authorization failed due to insufficient permissions.
    """
    reason = 'Authorization failed due to insufficient permissions.'


class NotFoundError(PlisioError):
    """
    404 Not Found.
    RESOURCE_NOT_FOUND.
    The specified resource does not exist.
    """
    reason = 'The specified resource does not exist.'


class MethodNotAllowedError(PlisioError):
    """
    405 Method Not Allowed.
    METHOD_NOT_SUPPORTED.
    The server does not implement the requested HTTP method.
    """
    reason = 'The server does not implement the requested HTTP method.'


class NotAcceptableError(PlisioError):
    """
    406 Not Acceptable.
    MEDIA_TYPE_NOT_ACCEPTABLE.
    The server does not implement the media type that would be acceptable to the client.
    """
    reason = 'The server does not implement the media type that would be ' \
             'acceptable to the client.'


class UnsupportedMediaTypeError(PlisioError):
    """
    415 Unsupported Media Type.
    UNSUPPORTED_MEDIA_TYPE.
    The server does not support the request payload’s media type.
    """
    reason = 'The server does not support the request payload’s media type.'


class UnprocessableEntityTypeError(PlisioError):
    """
    422 Unprocessable Entity.
    UNPROCESSABLE_ENTITY.
    The API cannot complete the requested action, or the request action is
    semantically incorrect or fails business validation.
    """
    reason = 'The API cannot complete the requested action, or the request ' \
             'action is semantically incorrect or fails business validation.'


class RateLimitReachedError(PlisioError):
    """
    429 Rate Limit Reached.
    RATE_LIMIT_REACHED.
    Too many requests. Blocked due to rate limiting.
    """
    reason = 'Too many requests. Blocked due to rate limiting.'


class InternalServerError(PlisioError):
    """
    500 Internal Server Error.
    INTERNAL_SERVER_ERROR.
    An internal server error has occurred.
    """
    reason = 'An internal server error has occurred.'


class ServiceUnavailableError(PlisioError):
    """
    503 Service Unavailable.
    SERVICE_UNAVAILABLE.
    Service Unavailable.
    """
    reason = 'Service Unavailable.'



from typing import Type, Dict, Optional, Union, List, Any

import aiohttp
import requests
import json
import hmac
import hashlib

import plisio


class _PlisioUrl:
    balance = 'balances'
    currencies = 'currencies'
    invoice = 'invoices/new'
    commission = 'operations/commission'
    withdraw = 'operations/withdraw'
    fee = 'operations/fee'
    fee_plan = 'operations/fee-plan'
    operations = 'operations'


class _PlisioRequest:
    def __init__(
            self,
            url: str,
            data: Dict[str, Any],
            response_class: Type['plisio.PlisioModel'],
            method: str = 'get',
    ):
        self.url = url
        self.data = data
        self.__response_class = response_class
        self.method = method

        self.__processed = False
        self.__response_status = None
        self.__response = None

    def set_response(self, status_code: int, response_dict: 'plisio.RType'):
        if self.__processed:
            raise plisio.RequestAlreadyProcessed()
        self.__processed = True
        self.__response_status = status_code
        self.__handle_processed_request(response_dict)

    def __handle_processed_request(self, response_dict: 'plisio.RType') -> 'plisio.ModelType':
        if self.response_status in [200, 201]:
            if self.__response_class is not None:
                data: 'plisio.RType' = response_dict['data']
                if isinstance(data, list):
                    self.__response = self.__response_class.list_of_models(data)
                else:
                    self.__response = self.__response_class.from_response(data)
        else:
            self.__handle_exception(response_dict)

    def __handle_exception(self, response_dict: 'plisio.RType'):
        message = response_dict.get('message')
        if self.response_status == 400:
            raise plisio.BadRequestError(message)
        if self.response_status == 401:
            raise plisio.UnauthorizedError(message)
        if self.response_status == 403:
            raise plisio.ForbiddenError(message)
        if self.response_status == 404:
            raise plisio.NotFoundError(message)
        if self.response_status == 405:
            raise plisio.MethodNotAllowedError(message)
        if self.response_status == 406:
            raise plisio.NotAcceptableError(message)
        if self.response_status == 415:
            raise plisio.UnsupportedMediaTypeError(message)
        if self.response_status == 422:
            raise plisio.UnprocessableEntityTypeError(message)
        if self.response_status == 429:
            raise plisio.RateLimitReachedError(message)
        if self.response_status == 500:
            raise plisio.InternalServerError(message)
        if self.response_status == 503:
            raise plisio.ServiceUnavailableError(message)
        raise plisio.PlisioError(message)

    @property
    def response_status(self) -> int:
        if self.__processed:
            return self.__response_status
        raise plisio.RequestNotProcessed()

    @property
    def response(self) -> 'plisio.ModelType':
        if self.__processed:
            return self.__response
        raise plisio.RequestNotProcessed()


class _BaseClient:
    _url = _PlisioUrl
    __api_url = 'https://api.plisio.net/api/v1/'

    def __init__(self, api_key: str):
        self.__api_key = api_key

    @staticmethod
    def __prepare_data(data: Dict[str, Any]):
        for key in list(data):
            if data[key] is None:
                del data[key]
            elif isinstance(data[key], float):
                data[key] = "{:.8f}".format(data[key]).rstrip('0')
            elif isinstance(data[key], bool):
                if data[key]:
                    data[key] = '1'
                else:
                    del data[key]

    def __create_request(
            self,
            uri: str,
            data: Dict[str, Any],
            response_class: Type['plisio.PlisioModel'],
            method: str = 'get',
    ) -> '_PlisioRequest':
        url = self.__api_url + uri
        self.__prepare_data(data)
        data['api_key'] = self.__api_key
        return _PlisioRequest(url, data, response_class, method)

    def _get_balance_request(self, **kwargs) -> '_PlisioRequest':
        return self.__create_request(
            self._url.balance + '/' + kwargs['currency'].name,
            {},
            plisio.Balance
        )

    def _get_currencies_request(self, **kwargs) -> '_PlisioRequest':
        url_ = '/' + kwargs['fiat_currency'].name if kwargs.get('fiat_currency') else ''
        return self.__create_request(
            self._url.currencies + url_,
            {},
            plisio.Currency
        )

    def _invoice_request(self, **kwargs) -> '_PlisioRequest':
        return self.__create_request(
            self._url.invoice,
            {
                'currency': kwargs['currency'].name,
                'order_name': kwargs['order_name'],
                'order_number': kwargs['order_number'],
                'amount': kwargs['amount'],
                'source_currency': kwargs['source_currency'] and kwargs['source_currency'].name,
                'source_amount': kwargs['source_amount'],
                'allowed_psys_cids': kwargs['allowed_currencies'] and ','.join(
                    [c.name for c in kwargs['allowed_currencies']]
                ),
                'description': kwargs['description'],
                'callback_url': kwargs['callback_url'],
                'email': kwargs['email'],
                'language': kwargs['language'],
                'plugin': kwargs['plugin'],
                'version': kwargs['version'],
                'redirect_to_invoice': kwargs['redirect_to_invoice'],
                'expire_min': kwargs['expire_min'],
            },
            plisio.Invoice
        )

    def _get_commission_request(self, **kwargs) -> '_PlisioRequest':
        return self.__create_request(
            self._url.commission + '/' + kwargs['crypto_currency'].name,
            {
                'addresses': kwargs['addresses'] and ','.join(kwargs['addresses']),
                'amounts':
                    ','.join(map(lambda s: "{:.8f}".format(s), kwargs['amounts']))
                    if isinstance(kwargs['amounts'], list)
                    else kwargs['amounts'] and "{:.8f}".format(kwargs['amounts']),
                'type': kwargs['type_'] and kwargs['type_'].name,
                'feePlan': kwargs['fee_plan'] and kwargs['fee_plan'].name,
                'customFeeRate': kwargs['custom_fee_rate'],
            },
            plisio.Commission
        )

    def _withdraw_request(self, **kwargs) -> '_PlisioRequest':
        return self.__create_request(
            self._url.withdraw,
            {
                'psys_cid': kwargs['crypto_currency'].name,
                'to': ','.join(kwargs['to']) if isinstance(kwargs['to'], list) else kwargs['to'],
                'amount':
                    ','.join(map(lambda s: "{:.8f}".format(s), kwargs['amount']))
                    if isinstance(kwargs['amount'], list)
                    else "{:.8f}".format(kwargs['amount']),
                'type': kwargs['type_'] and kwargs['type_'].name,
                'feePlan': kwargs['fee_plan'] and kwargs['fee_plan'].name,
                'feeRate': kwargs['fee_rate'],
            },
            plisio.Withdraw
        )

    def _get_fee_request(self, **kwargs) -> '_PlisioRequest':
        return self.__create_request(
            self._url.fee + '/' + kwargs['currency'].name,
            {
                'addresses':
                    ','.join(kwargs['addresses']) if isinstance(kwargs['addresses'], list) else kwargs['addresses'],
                'amounts':
                    ','.join(map(lambda s: "{:.8f}".format(s), kwargs['amounts']))
                    if isinstance(kwargs['amounts'], list)
                    else "{:.8f}".format(kwargs['amounts']),
                'feePlan': kwargs['fee_plan'] and kwargs['fee_plan'].name,
            },
            plisio.Fee
        )

    def _get_fee_plan_request(self, **kwargs) -> '_PlisioRequest':
        return self.__create_request(
            self._url.fee_plan + '/' + kwargs['currency'].name,
            {},
            plisio.FeePlan
        )

    def _get_operations_request(self, **kwargs) -> '_PlisioRequest':
        return self.__create_request(
            self._url.operations,
            {
                'page': kwargs['page'],
                'limit': kwargs['limit'],
                'shop_id': kwargs['shop_id'],
                'type': kwargs['type_'] and plisio.OperationType[kwargs['type_']],
                'status': kwargs['status'] and plisio.OperationStatus[kwargs['status']],
                'currency': kwargs['currency'] and plisio.CryptoCurrency[kwargs['currency']],
                'search': kwargs['search'],
            },
            plisio.Operations
        )

    def _get_operation_request(self, **kwargs) -> '_PlisioRequest':
        return self.__create_request(
            self._url.operations + '/' + kwargs['id_'],
            {},
            plisio.Operation
        )

    def validate_callback(self, data: str) -> bool:
        jsonObj = json.loads(data)
        verifyHash = jsonObj['verify_hash']
        del jsonObj["verify_hash"]
        json.dumps(jsonObj, sort_keys=True)
        key = bytes(str(self.__api_key), 'utf8')
        postStr = bytes(json.dumps(jsonObj, separators=(',', ':')), 'utf8');
        digester = hmac.new(key, postStr, hashlib.sha1)
        hash = digester.hexdigest()
        return hash == verifyHash


class PlisioClient(_BaseClient):
    @staticmethod
    def _send_request(request: '_PlisioRequest') -> 'plisio.ModelType':
        try:
            _req = requests.request(
                request.method,
                request.url,
                params=request.data,
            )
            status = _req.status_code
            if _req.history:
                data = {'status': 'redirect', 'data': {'invoice_url': _req.url}}
            else:
                data = _req.json()
        except requests.exceptions.RequestException as re:
            raise plisio.UnknownPlisioAPIError() from re
        else:
            request.set_response(status, data)
            return request.response

    def get_balance(self, currency: 'plisio.CryptoCurrency') -> 'plisio.Balance':
        """
        /balances/{psys_cid}
        Get cryptocurrency balance
        """
        request = self._get_balance_request(currency=currency)
        return self._send_request(request)

    def get_currencies(
            self,
            fiat_currency: Optional['plisio.FiatCurrency'] = None,
    ) -> List['plisio.Currency']:
        """
        /currencies/{fiat}
        List of supported cryptocurrencies
        """
        request = self._get_currencies_request(fiat_currency=fiat_currency)
        return self._send_request(request)


    def invoice(
            self,
            currency: 'plisio.CryptoCurrency',
            order_name: str,
            order_number: int,
            amount: Optional[float] = None,
            source_currency: Optional['plisio.FiatCurrency'] = None,
            source_amount: Optional[float] = None,
            allowed_currencies: Optional[List['plisio.CryptoCurrency']] = None,
            description: Optional[str] = None,
            callback_url: Optional[str] = None,
            email: Optional[str] = None,
            language: Optional[str] = 'en_US',
            plugin: Optional[str] = None,
            version: Optional[str] = None,
            redirect_to_invoice: Optional[bool] = None,
            expire_min: Optional[int] = None,
    ) -> List['plisio.Invoice']:
        """
        /invoices/new
        Create new invoice
        """
        request = self._invoice_request(
            currency=currency,
            order_name=order_name,
            order_number=order_number,
            amount=amount,
            source_currency=source_currency,
            source_amount=source_amount,
            allowed_currencies=allowed_currencies,
            description=description,
            callback_url=callback_url,
            email=email,
            language=language,
            plugin=plugin,
            version=version,
            redirect_to_invoice=redirect_to_invoice,
            expire_min=expire_min,
        )
        return self._send_request(request)

    create_invoice = invoice

    def get_commission(
            self,
            crypto_currency: 'plisio.CryptoCurrency',
            addresses: Optional[Union[str, List[str]]] = None,
            amounts: Optional[Union[float, List[float]]] = None,
            type_: Optional['plisio.OperationType'] = None,
            fee_plan: Optional['plisio.PlanName'] = None,
            custom_fee_rate: Optional[int] = None,
    ) -> 'plisio.Commission':
        """
        /operations/commission/{psys_cid}
        Estimate cryptocurrency fee and Plisio commission
        """
        request = self._get_commission_request(
            crypto_currency=crypto_currency,
            addresses=addresses,
            amounts=amounts,
            type_=type_,
            fee_plan=fee_plan,
            custom_fee_rate=custom_fee_rate,
        )
        return self._send_request(request)

    def withdraw(
            self,
            crypto_currency: 'plisio.CryptoCurrency',
            to: Union[str, List[str]],
            amount: Union[float, List[float]],
            type_: Optional['plisio.OperationType'] = None,
            fee_plan: Optional['plisio.PlanName'] = None,
            fee_rate: Optional[float] = None,
    ) -> 'plisio.Withdraw':
        """
        /operations/withdraw
        Withdrawal operation
        """
        request = self._withdraw_request(
            crypto_currency=crypto_currency,
            to=to,
            amount=amount,
            type_=type_,
            fee_plan=fee_plan,
            fee_rate=fee_rate,
        )
        return self._send_request(request)

    def get_fee(
            self,
            currency: 'plisio.CryptoCurrency',
            addresses: Union[str, List[str]],
            amounts: Union[float, List[float]],
            fee_plan: Optional['plisio.PlanName'] = None,
    ) -> 'plisio.Fee':
        """
        /operations/fee/{psys_cid}
        Estimate fee
        """
        request = self._get_fee_request(
            currency=currency,
            addresses=addresses,
            amounts=amounts,
            fee_plan=fee_plan,
        )
        return self._send_request(request)

    def get_fee_plan(self, currency: 'plisio.CryptoCurrency') -> 'plisio.FeePlan':
        """
        /operations/fee-plan/{psys_cid}
        Get Plisio fee plans
        """
        request = self._get_fee_plan_request(
            currency=currency,
        )
        return self._send_request(request)

    def get_operations(
            self,
            page: Optional[int] = None,
            limit: Optional[int] = None,
            shop_id: Optional[str] = None,
            type_: Optional['plisio.OperationType'] = None,
            status: Optional['plisio.OperationStatus'] = None,
            currency: Optional['plisio.CryptoCurrency'] = None,
            search: Optional[str] = None,
    ) -> 'plisio.Operations':
        """
        /operations
        List of all user transactions
        """
        request = self._get_operations_request(
            page=page,
            limit=limit,
            shop_id=shop_id,
            type_=type_,
            status=status,
            currency=currency,
            search=search,
        )
        return self._send_request(request)

    def get_operation(self, id_: str) -> 'plisio.Operation':
        """
        /operations/{id}
        Transaction details
        """
        request = self._get_operation_request(
            id_=id_,
        )
        return self._send_request(request)


class PlisioAioClient(_BaseClient):
    @staticmethod
    async def _send_request(request: '_PlisioRequest'):
        try:
            async with aiohttp.ClientSession() as session:
                async with session.request(
                        request.method,
                        request.url,
                        params=request.data,
                ) as _req:
                    status = _req.status
                    if _req.history:
                        data = {'status': 'redirect', 'data': {'invoice_url': _req.url}}
                    else:
                        data = await _req.json()
        except aiohttp.ClientError as ce:
            raise plisio.UnknownPlisioAPIError() from ce
        else:
            request.set_response(status, data)
            return request.response

    async def get_balance(
            self,
            currency: 'plisio.CryptoCurrency',
    ) -> 'plisio.Balance':
        """
        /balances/{psys_cid}
        Async method to get cryptocurrency balance
        """
        request = self._get_balance_request(
            currency=currency,
        )
        return await self._send_request(request)

    async def get_currencies(
            self,
            fiat_currency: Optional['plisio.FiatCurrency'] = None,
    ) -> List['plisio.Currency']:
        """
        /currencies/{fiat}
        List of supported cryptocurrencies
        """
        request = self._get_currencies_request(fiat_currency=fiat_currency)
        return await self._send_request(request)

    async def invoice(
            self,
            currency: 'plisio.CryptoCurrency',
            order_name: str,
            order_number: int,
            amount: float,
            source_currency: Optional['plisio.FiatCurrency'] = None,
            source_amount: Optional[float] = None,
            allowed_currencies: Optional[List['plisio.CryptoCurrency']] = None,
            description: Optional[str] = None,
            callback_url: Optional[str] = None,
            email: Optional[str] = None,
            language: Optional[str] = 'en_US',
            plugin: Optional[str] = None,
            version: Optional[str] = None,
            redirect_to_invoice: Optional[bool] = None,
            expire_min: Optional[int] = None,
    ) -> List['plisio.Invoice']:
        """
        /invoices/new
        Async method to create new invoice
        """
        request = self._invoice_request(
            currency=currency,
            order_name=order_name,
            order_number=order_number,
            amount=amount,
            source_currency=source_currency,
            source_amount=source_amount,
            allowed_currencies=allowed_currencies,
            description=description,
            callback_url=callback_url,
            email=email,
            language=language,
            plugin=plugin,
            version=version,
            redirect_to_invoice=redirect_to_invoice,
            expire_min=expire_min,
        )
        return await self._send_request(request)

    async def get_commission(
            self,
            crypto_currency: 'plisio.CryptoCurrency',
            addresses: Optional[Union[str, List[str]]] = None,
            amounts: Optional[Union[float, List[float]]] = None,
            type_: Optional['plisio.OperationType'] = None,
            fee_plan: Optional['plisio.PlanName'] = None,
            custom_fee_rate: Optional[int] = None,
    ) -> 'plisio.Commission':
        """
        /operations/commission/{psys_cid}
        Async method to estimate cryptocurrency fee and Plisio commission
        """
        request = self._get_commission_request(
            crypto_currency=crypto_currency,
            addresses=addresses,
            amounts=amounts,
            type_=type_,
            fee_plan=fee_plan,
            custom_fee_rate=custom_fee_rate,
        )
        return await self._send_request(request)

    async def withdraw(
            self,
            crypto_currency: 'plisio.CryptoCurrency',
            to: Union[str, List[str]],
            amount: Union[float, List[float]],
            fee_plan: Optional['plisio.PlanName'] = None,
            fee_rate: Optional[float] = None,
            type_: Optional['plisio.OperationType'] = None,
    ) -> 'plisio.Withdraw':
        """
        /operations/withdraw
        Async method to withdrawal operation
        """
        request = self._withdraw_request(
            crypto_currency=crypto_currency,
            to=to,
            amount=amount,
            fee_plan=fee_plan,
            fee_rate=fee_rate,
            type_=type_,
        )
        return await self._send_request(request)

    async def get_fee(
            self,
            currency: 'plisio.CryptoCurrency',
            addresses: Union[str, List[str]],
            amounts: Union[float, List[float]],
            fee_plan: Optional['plisio.PlanName'] = None,
    ) -> 'plisio.Fee':
        """
        /operations/fee/{psys_cid}
        Async method to estimate fee
        """
        request = self._get_fee_request(
            currency=currency,
            addresses=addresses,
            amounts=amounts,
            fee_plan=fee_plan,
        )
        return await self._send_request(request)

    async def get_fee_plan(self, currency: 'plisio.CryptoCurrency') -> 'plisio.FeePlan':
        """
        /operations/fee-plan/{psys_cid}
        Async method to get Plisio fee plans
        """
        request = self._get_fee_plan_request(
            currency=currency,
        )
        return await self._send_request(request)

    async def get_operations(
            self,
            page: Optional[int] = None,
            limit: Optional[int] = None,
            shop_id: Optional[str] = None,
            type_: Optional['plisio.OperationType'] = None,
            status: Optional['plisio.OperationStatus'] = None,
            currency: Optional['plisio.CryptoCurrency'] = None,
            search: Optional[str] = None,
    ) -> List['plisio.Operation']:
        """
        /operations
        Async method to list of all user transactions
        """
        request = self._get_operations_request(
            page=page,
            limit=limit,
            shop_id=shop_id,
            type_=type_,
            status=status,
            currency=currency,
            search=search,
        )
        return await self._send_request(request)

    async def get_operation(self, id_: str) -> 'plisio.Operation':
        """
        /operations/{id}
        Async method to transaction details
        """
        request = self._get_operation_request(
            id_=id_,
        )
        return await self._send_request(request)



from typing import Dict, List, Optional, Union

import plisio


class PlisioModel:
    @classmethod
    def from_response(cls, response_dict: 'plisio.RType'):
        raise NotImplementedError()

    @classmethod
    def list_of_models(cls, response_list: List['plisio.RType']):
        return [
            cls.from_response(response_dict)
            for response_dict in response_list
        ]

    def __repr__(self):
        return '<' + f'{super().__repr__()}: ' + str(self.__dict__) + '>'


class Balance(PlisioModel):
    """
    /balances/{psys_cid}
    Get plisio.CryptoCurrency balance
    """

    def __init__(
            self,
            currency: 'plisio.CryptoCurrency',
            balance: float,
            locked_balance: Optional[float]
    ):
        self.currency = currency
        self.balance = balance
        self.locked_balance = locked_balance

    @classmethod
    def from_response(cls, response_dict: 'plisio.RType') -> 'Balance':
        return cls(
            response_dict.get('psys_cid') and plisio.CryptoCurrency[response_dict['psys_cid']],
            response_dict.get('balance') and float(response_dict['balance']),
            response_dict.get('lockedBalance') and float(response_dict['lockedBalance']),
        )


class Currency(PlisioModel):
    """
    /currencies/{fiat}
    List of supported cryptocurrencies
    """

    def __init__(
            self,
            currency: 'plisio.CryptoCurrency',
            icon: str,
            rate_usd: float,
            price_usd: float,
            precision: int,
            fiat: 'plisio.FiatCurrency',
            fiat_rate: float,
            min_sum_in: float,
            invoice_commission_percentage: int,
    ):
        self.currency = currency
        self.icon = icon
        self.rate_usd = rate_usd
        self.price_usd = price_usd
        self.precision = precision
        self.fiat = fiat
        self.fiat_rate = fiat_rate
        self.min_sum_in = min_sum_in
        self.invoice_commission_percentage = invoice_commission_percentage

    @classmethod
    def from_response(cls, response_dict: 'plisio.RType') -> 'Currency':
        return cls(
            response_dict.get('cid') and plisio.CryptoCurrency[response_dict['cid'].replace('-', '_')],
            response_dict.get('icon') and str(response_dict['icon']),
            response_dict.get('rate_usd') and float(response_dict['rate_usd']),
            response_dict.get('price_usd') and float(response_dict['price_usd']),
            response_dict.get('precision') and int(response_dict['precision']),
            response_dict.get('fiat') and plisio.FiatCurrency[response_dict['fiat']],
            response_dict.get('fiat_rate') and float(response_dict['fiat_rate']),
            response_dict.get('min_sum_in') and float(response_dict['min_sum_in']),
            response_dict.get('invoice_commission_percentage') and int(response_dict['invoice_commission_percentage']),
        )


class Invoice(PlisioModel):
    """
    /invoices/new
    Create new invoice
    """

    def __init__(
            self,
            txn_id: str,
            invoice_url: str,
            amount: Optional[float],
            pending_amount: Optional[float],
            wallet_hash: Optional[str],
            currency: Optional['plisio.CryptoCurrency'],
            source_currency: Optional['plisio.FiatCurrency'],
            source_rate: Optional[float],
            expected_confirmations: Optional[int],
            qr_code: Optional[str],
            verify_hash: Optional[str],
            invoice_commission: Optional[float],
            invoice_sum: Optional[float],
            invoice_total_sum: Optional[float],
    ):
        self.txn_id = txn_id
        self.invoice_url = invoice_url
        self.amount = amount
        self.pending_amount = pending_amount
        self.wallet_hash = wallet_hash
        self.currency = currency
        self.source_currency = source_currency
        self.source_rate = source_rate
        self.expected_confirmations = expected_confirmations
        self.qr_code = qr_code
        self.verify_hash = verify_hash
        self.invoice_commission = invoice_commission
        self.invoice_sum = invoice_sum
        self.invoice_total_sum = invoice_total_sum

    @classmethod
    def from_response(cls, response_dict: 'plisio.RType') -> 'Invoice':
        return cls(
            response_dict.get('txn_id') and str(response_dict['txn_id']),
            response_dict.get('invoice_url') and str(response_dict['invoice_url']),
            response_dict.get('amount') and float(response_dict['amount']),
            response_dict.get('pending_amount') and float(response_dict['pending_amount']),
            response_dict.get('wallet_hash') and str(response_dict['wallet_hash']),
            response_dict.get('currency') and plisio.CryptoCurrency[response_dict.get('currency')],
            response_dict.get('source_currency') and plisio.FiatCurrency[response_dict.get('source_currency')],
            response_dict.get('source_rate') and float(response_dict['source_rate']),
            response_dict.get('expected_confirmations') and int(response_dict['expected_confirmations']),
            response_dict.get('qr_code') and str(response_dict['qr_code']),
            response_dict.get('verify_hash') and str(response_dict['verify_hash']),
            response_dict.get('invoice_commission') and float(response_dict['invoice_commission']),
            response_dict.get('invoice_sum') and float(response_dict['invoice_sum']),
            response_dict.get('invoice_total_sum') and float(response_dict['invoice_total_sum']),
        )


class Plan(PlisioModel):
    """
    Sub-model for FeePlan, WithdrawParams
    """

    def __init__(
            self,
            conf_target: int,
            fee_rate: int,
            dynamic_field: str,
            plan: 'plisio.PlanName',
            unit: str,
            value: float,
    ):
        self.conf_target = conf_target
        self.fee_rate = fee_rate
        self.dynamic_field = dynamic_field
        self.plan = plan
        self.unit = unit
        self.value = value

    @classmethod
    def from_response(cls, response_dict: Dict[str, Union[str, dict]]) -> 'Plan':
        return cls(
            response_dict.get('conf_target') and int(response_dict['conf_target']),
            response_dict.get('feeRate') and int(response_dict['feeRate']),
            response_dict.get('dynamicField'),
            response_dict.get('plan') and plisio.PlanName[response_dict['plan']],
            response_dict.get('unit'),
            response_dict.get('value') and float(response_dict['value']),
        )


class FeePlan(PlisioModel):
    """
    /operations/fee-plan/{psys_cid}
    Get Plisio fee plans
    Sub-model for Commission
    """

    def __init__(
            self,
            currency: 'plisio.CryptoCurrency',
            economy: Optional['Plan'] = None,
            normal: Optional['Plan'] = None,
            priority: Optional['Plan'] = None,
            custom: Optional['Plan'] = None,
    ):
        self.currency = currency
        self.economy = economy
        self.normal = normal
        self.priority = priority
        self.custom = custom

    @classmethod
    def from_response(cls, response_dict: Dict[str, Union[str, dict]]) -> 'FeePlan':
        return cls(
            response_dict.get('psys_cid') and plisio.CryptoCurrency[response_dict['psys_cid']],
            response_dict.get('economy') and Plan.from_response(response_dict['economy']),
            response_dict.get('normal') and Plan.from_response(response_dict['normal']),
            response_dict.get('priority') and Plan.from_response(response_dict['priority']),
            response_dict.get('custom') and Plan.from_response(response_dict['custom']),
        )


class Custom(PlisioModel):
    """
    Sub-model for Commission
    """

    def __init__(
            self,
            min_: int,
            max_: int,
            default: int,
            borders: int,
            unit: str,
    ):
        self.min = min_
        self.max = max_
        self.default = default
        self.borders = borders
        self.unit = unit

    @classmethod
    def from_response(cls, response_dict: Dict[str, Union[str, dict]]) -> 'Custom':
        return cls(
            response_dict.get('min') and int(response_dict['min']),
            response_dict.get('max') and int(response_dict['max']),
            response_dict.get('default') and int(response_dict['default']),
            response_dict.get('borders') and int(response_dict['borders']),
            response_dict.get('unit'),
        )


class Commission(PlisioModel):
    """
    /operations/commission/{psys_cid}
    Estimate cryptocurrency fee and Plisio commission
    """

    def __init__(
            self,
            commission: float,
            fee: float,
            max_amount: float,
            plan: 'plisio.PlanName',
            use_wallet: Optional[int],
            use_wallet_balance: Optional[int],
            plans: 'FeePlan',
            custom: 'Custom',
            errors: Optional[int],
            custom_fee_rate: int
    ):
        self.commission = commission
        self.fee = fee
        self.max_amount = max_amount
        self.plan = plan
        self.use_wallet = use_wallet
        self.use_wallet_balance = use_wallet_balance
        self.plans = plans
        self.custom = custom
        self.errors = errors
        self.custom_fee_rate = custom_fee_rate

    @classmethod
    def from_response(cls, response_dict: Dict[str, Union[str, dict]]) -> 'Commission':
        return cls(
            response_dict.get('commission') and float(response_dict['commission']),
            response_dict.get('fee') and float(response_dict['fee']),
            response_dict.get('maxAmount') and float(response_dict['maxAmount']),
            response_dict.get('plan') and plisio.PlanName[response_dict['plan']],
            response_dict.get('useWallet') and int(response_dict['useWallet']),
            response_dict.get('useWalletBalance') and int(response_dict['useWalletBalance']),
            response_dict.get('plans') and FeePlan.from_response(response_dict['plans']),
            response_dict.get('custom') and Custom.from_response(response_dict['custom']),
            response_dict.get('errors') and int(response_dict['errors']),
            response_dict.get('customFeeRate') and int(response_dict['customFeeRate']),
        )


class WithdrawParams(PlisioModel):
    """
    Sub-model for Withdraw
    """

    def __init__(
            self,
            source_currency: 'plisio.CryptoCurrency',
            source_rate: float,
            usd_rate: Optional[float],
            fee: Optional['Plan'],
    ):
        self.source_currency = source_currency
        self.source_rate = source_rate
        self.usd_rate = usd_rate
        self.fee = fee

    @classmethod
    def from_response(cls, response_dict: 'plisio.RType') -> 'WithdrawParams':
        return cls(
            response_dict.get('source_currency'),
            response_dict.get('source_rate') and float(response_dict['source_rate']),
            response_dict.get('usd_rate') and float(response_dict['usd_rate']),
            response_dict.get('fee') and Plan.from_response(response_dict['fee']),
        )


class Withdraw(PlisioModel):
    """
    /operations/withdraw
    Create a copy of invoice
    """

    def __init__(
            self,
            type_: 'plisio.OperationType',
            status: str,
            currency: 'plisio.CryptoCurrency',
            source_currency: 'plisio.FiatCurrency',
            source_rate: float,
            fee: float,
            wallet_hash: str,
            sendmany: List[Dict[str, float]],
            params: 'WithdrawParams',
            created_at_utc: int,
            amount: float,
            tx_url: str,
            tx_id: List[str],
            id_: str,
    ):
        self.type = type_
        self.status = status
        self.currency = currency
        self.source_currency = source_currency
        self.source_rate = source_rate
        self.fee = fee
        self.wallet_hash = wallet_hash
        self.sendmany = sendmany
        self.params = params
        self.created_at_utc = created_at_utc
        self.amount = amount
        self.tx_url = tx_url
        self.tx_id = tx_id
        self.id = id_

    @classmethod
    def from_response(cls, response_dict: 'plisio.RType') -> 'Withdraw':
        return cls(
            response_dict.get('type') and plisio.OperationType[response_dict['type']],
            response_dict.get('status') and str(response_dict['status']),
            response_dict.get('psys_cid') and plisio.CryptoCurrency[response_dict['psys_cid']],
            response_dict.get('source_currency') and plisio.FiatCurrency[response_dict['source_currency']],
            response_dict.get('source_rate') and float(response_dict['source_rate']),
            response_dict.get('fee') and float(response_dict['fee']),
            response_dict.get('wallet_hash'),
            response_dict.get('sendmany') and [[{k: float(v)} for k, v in sm.items()] for sm in response_dict['sendmany']],
            response_dict.get('params') and WithdrawParams.from_response(response_dict['params']),
            response_dict.get('created_at_utc') and int(response_dict['created_at_utc']),
            response_dict.get('amount') and float(response_dict['amount']),
            response_dict.get('tx_url') and str(response_dict['tx_url']),
            response_dict.get('tx_id'),
            response_dict.get('id'),
        )


class Fee(PlisioModel):
    """
    /operations/fee/{psys_cid}
    Estimate fee
    """

    def __init__(
            self,
            fee: float,
            currency: 'plisio.CryptoCurrency',
            plan: 'plisio.PlanName'
    ):
        self.fee = fee
        self.currency = currency
        self.plan = plan

    @classmethod
    def from_response(cls, response_dict: 'plisio.RType') -> 'Fee':
        return cls(
            response_dict.get('fee') and float(response_dict['fee']),
            response_dict.get('psys_cid') and plisio.CryptoCurrency[response_dict['psys_cid']],
            response_dict.get('plan') and plisio.PlanName[response_dict['plan']],
        )


class OperationTx(PlisioModel):
    """
    Sub-model for Operation
    """

    def __init__(
            self,
            txid: str,
            block: int,
            confirmations: int,
            value: float,
            processed: bool,
            fail_retry: int,
            fee_rate: float,
            fee_rate_unit: str,
            url: str,
            wallet_hash: List[str],
    ):
        self.txid = txid
        self.block = block
        self.confirmations = confirmations
        self.value = value
        self.processed = processed
        self.fail_retry = fail_retry
        self.fee_rate = fee_rate
        self.fee_rate_unit = fee_rate_unit
        self.url = url
        self.wallet_hash = wallet_hash

    @classmethod
    def from_response(cls, response_dict: 'plisio.RType') -> 'OperationTx':
        return cls(
            response_dict.get('txid') and str(response_dict['txid']),
            response_dict.get('block') and int(response_dict['block']),
            response_dict.get('confirmations') and int(response_dict['confirmations']),
            response_dict.get('value') and float(response_dict['value']),
            response_dict.get('processed'),
            response_dict.get('failRetry') and int(response_dict['failRetry']),
            response_dict.get('feeRate') and float(response_dict['feeRate']),
            response_dict.get('feeRateUnit') and response_dict['feeRateUnit'],
            response_dict.get('url'),
            response_dict.get('wallet_hash'),
        )


class OperationParams(PlisioModel):
    """
    Sub-model for Operation
    """

    def __init__(
            self,
            order_number: str,
            order_name: Optional[str],
            source_amount: Optional[float],
            source_currency: 'plisio.CryptoCurrency',
            currency: Optional['plisio.CryptoCurrency'],
            amount: Optional[float],
            source_rate: float,
            email: str,
            usd_rate: Optional[float],
            fee: Optional['Plan'],
    ):
        self.source_currency = source_currency
        self.source_rate = source_rate
        self.usd_rate = usd_rate
        self.fee = fee
        self.order_number = order_number
        self.order_name = order_name
        self.source_amount = source_amount
        self.currency = currency
        self.amount = amount
        self.email = email

    @classmethod
    def from_response(cls, response_dict: 'plisio.RType') -> 'OperationParams':
        return cls(
            response_dict.get('order_number'),
            response_dict.get('order_name'),
            response_dict.get('source_amount') and float(response_dict['source_amount']),
            response_dict.get('source_currency'),
            response_dict.get('currency'),
            response_dict.get('amount') and float(response_dict['amount']),
            response_dict.get('source_rate') and float(response_dict['source_rate']),
            response_dict.get('email'),
            response_dict.get('usd_rate') and float(response_dict['usd_rate']),
            response_dict.get('fee') and Plan.from_response(response_dict['fee']),
        )


class Operation(PlisioModel):
    """
    /operations
    List of all user transactions
    /operations/{id}
    Create new invoice
    """

    def __init__(
            self,
            user_id: int,
            shop_id: str,
            type_: 'plisio.OperationType',
            status: 'plisio.OperationStatus',
            pending_sum: float,
            currency: 'plisio.CryptoCurrency',
            source_currency: 'plisio.FiatCurrency',
            source_rate: float,
            fee: float,
            wallet_hash: str,
            sendmany: List[Dict[str, float]],
            params: 'OperationParams',
            expire_at_utc: int,
            created_at_utc: int,
            amount: float,
            sum_: float,
            commission: float,
            tx_url: str,
            tx_id: List[str],
            id_: str,
            actual_sum: float,
            actual_commission: float,
            actual_fee: float,
            actual_invoice_sum: float,
            tx: List[OperationTx],
            status_code: int,
    ):
        self.user_id = user_id
        self.shop_id = shop_id
        self.type = type_
        self.status = status
        self.pending_sum = pending_sum
        self.currency = currency
        self.source_currency = source_currency
        self.source_rate = source_rate
        self.fee = fee
        self.wallet_hash = wallet_hash
        self.sendmany = sendmany
        self.params = params
        self.expire_at_utc = expire_at_utc
        self.created_at_utc = created_at_utc
        self.amount = amount
        self.sum = sum_
        self.commission = commission
        self.tx_url = tx_url
        self.tx_id = tx_id
        self.id = id_
        self.actual_sum = actual_sum
        self.actual_commission = actual_commission
        self.actual_fee = actual_fee
        self.actual_invoice_sum = actual_invoice_sum
        self.tx = tx
        self.status_code = status_code

    @classmethod
    def from_response(cls, response_dict: 'plisio.RType') -> 'Operation':
        return cls(
            response_dict.get('user_id') and int(response_dict['user_id']),
            response_dict.get('shop_id'),
            response_dict.get('type') and plisio.OperationType[response_dict['type']],
            response_dict.get('status') and plisio.OperationStatus[response_dict['status']],
            response_dict.get('pending_sum') and float(response_dict['pending_sum']),
            response_dict.get('currency') and plisio.CryptoCurrency[response_dict['currency']],
            response_dict.get('source_currency') and plisio.FiatCurrency[response_dict['source_currency']],
            response_dict.get('source_rate') and float(response_dict['source_rate']),
            response_dict.get('fee') and float(response_dict['fee']),
            response_dict.get('wallet_hash'),
            response_dict.get('sendmany') and [[{k: float(v)} for k, v in sm.items()] for sm in response_dict['sendmany']],
            response_dict.get('params') and OperationParams.from_response(response_dict['params']),
            response_dict.get('expire_at_utc') and int(response_dict['expire_at_utc']),
            response_dict.get('created_at_utc') and int(response_dict['created_at_utc']),
            response_dict.get('amount') and float(response_dict['amount']),
            response_dict.get('sum') and float(response_dict['sum']),
            response_dict.get('commission') and float(response_dict['commission']),
            response_dict.get('tx_url'),
            response_dict.get('tx_id'),
            response_dict.get('id'),
            response_dict.get('actual_sum') and float(response_dict['actual_sum']),
            response_dict.get('actual_commission') and float(response_dict['actual_commission']),
            response_dict.get('actual_fee') and float(response_dict['actual_fee']),
            response_dict.get('actual_invoice_sum') and float(response_dict['actual_invoice_sum']),
            response_dict.get('tx') and [OperationTx.from_response(tx) for tx in response_dict['tx']],
            response_dict.get('status_code') and int(response_dict['status_code']),
        )


class Operations(PlisioModel):
    """
    Wrapper for Operation
    """

    def __init__(
            self,
            operations: List['Operation'],
            links: Dict[str, Dict[str, str]],
            meta: Dict[str, int],
    ):
        self.operations = operations
        self.links = links
        self.meta = meta

    @classmethod
    def from_response(cls, response_dict: 'plisio.RType') -> 'Operations':
        return cls(
            response_dict.get('operations') and [Operation.from_response(op) for op in response_dict['operations']],
            response_dict.get('_links'),
            response_dict.get('_meta'),
        )



from setuptools import setup


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='plisio',
    packages=['plisio'],
    version='1.0.8',
    license='MIT',
    description='Official Python SDK for Plisio API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Plisio',
    author_email='support@plisio.net',
    url='https://github.com/Plisio/plisio-python',
    download_url='https://github.com/Plisio/plisio-python',
    keywords=['plisio', 'crypto payment', 'sdk', 'bitcoin', 'etherium', 'crypto', 'blockchain'],
    install_requires=[
        'aiohttp',
        'requests',
        'hashlib',
        'hmac'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)



from enum import Enum, EnumMeta, auto


class _EnumMeta(EnumMeta):
    def __getitem__(self, item):
        if item is None:
            return None
        return super().__getitem__(item)


class CryptoCurrency(Enum, metaclass=_EnumMeta):
    ETH = 'Ethereum'
    BTC = 'Bitcoin'
    LTC = 'Litecoin'
    DASH = 'Dash'
    ZEC = 'Zcash'
    DOGE = 'Dogecoin'
    BCH = 'Bitcoin Cash'
    XMR = 'Monero'
    USDT = 'Tether ERC-20'
    SHIB = 'Shiba Inu',
    USDC = 'USD Coin ERC-20',
    TRX = 'Tron',
    USDT_TRX = 'Tether TRC-20',
    BNB = 'BNB Chain',
    USDT_BSC = 'Teather BEP-20',
    BUSD = 'Binance USD BEP-20',
    ETC = 'Ethereum Classic',
    TZEC = 'Zcash',
    BTT_TRX = 'BitTorrent-Chain TRC-20',
    TUSD = 'TrueUSD ERC-20'


class FiatCurrency(Enum, metaclass=_EnumMeta):
    AED = 'United Arab Emirates Dirham'
    AFN = 'Afghan Afghani'
    ALL = 'Albanian Lek'
    AMD = 'Armenian Dram'
    ANG = 'Netherlands Antillean Guilder'
    AOA = 'Angolan Kwanza'
    ARS = 'Argentine Peso'
    AUD = 'Australian Dollar'
    AWG = 'Aruban Florin'
    AZN = 'Azerbaijani Manat'
    BAM = 'Bosnia-Herzegovina Convertible Mark'
    BBD = 'Barbadian Dollar'
    BDT = 'Bangladeshi Taka'
    BGN = 'Bulgarian Lev'
    BHD = 'Bahraini Dinar'
    BIF = 'Burundian Franc'
    BMD = 'Bermuda Dollar'
    BND = 'Brunei Dollar'
    BOB = 'Bolivian Boliviano'
    BRL = 'Brazilian Real'
    BSD = 'Bahamian Dollar'
    BTN = 'Bhutanese Ngultrum'
    BWP = 'Botswana Pula'
    BYN = 'New Belarusian Ruble'
    BYR = 'Belarusian Ruble'
    BZD = 'Belize Dollar'
    CAD = 'Canadian Dollar'
    CDF = 'Congolese Franc'
    CHF = 'Swiss Franc'
    CLF = 'Chilean Unit of Account (UF)'
    CLP = 'Chilean Peso'
    CNY = 'Chinese Yuan'
    COP = 'Colombian Peso'
    CRC = 'Costa Rican Colón'
    CUC = 'Cuban Convertible Peso'
    CUP = 'Cuban Peso'
    CVE = 'Cape Verdean Escudo'
    CZK = 'Czech Republic Koruna'
    DJF = 'Djiboutian Franc'
    DKK = 'Danish Krone'
    DOP = 'Dominican Peso'
    DZD = 'Algerian Dinar'
    EGP = 'Egyptian Pound'
    ERN = 'Eritrean Nakfa'
    ETB = 'Ethiopian Birr'
    EUR = 'Euro'
    FJD = 'Fijian Dollar'
    FKP = 'Falkland Islands Pound'
    GBP = 'British Pound Sterling'
    GEL = 'Georgian Lari'
    GGP = 'Guernsey Pound'
    GHS = 'Ghanaian Cedi'
    GIP = 'Gibraltar Pound'
    GMD = 'Gambian Dalasi'
    GNF = 'Guinean Franc'
    GTQ = 'Guatemalan Quetzal'
    GYD = 'Guyanese Dollar'
    HKD = 'Hong Kong Dollar'
    HNL = 'Honduran Lempira'
    HRK = 'Croatian Kuna'
    HTG = 'Haitian Gourde'
    HUF = 'Hungarian Forint'
    IDR = 'Indonesian Rupiah'
    ILS = 'Israeli New Sheqel'
    IMP = 'Manx pound'
    INR = 'Indian Rupee'
    IQD = 'Iraqi Dinar'
    IRR = 'Iranian Rial'
    ISK = 'Icelandic Króna'
    JEP = 'Jersey Pound'
    JMD = 'Jamaican Dollar'
    JOD = 'Jordanian Dinar'
    JPY = 'Japanese Yen'
    KES = 'Kenyan Shilling'
    KGS = 'Kyrgyzstan Som'
    KHR = 'Cambodian Riel'
    KMF = 'Comorian Franc'
    KPW = 'North Korean Won'
    KRW = 'South Korean Won'
    KWD = 'Kuwaiti Dinar'
    KYD = 'Cayman Islands Dollar'
    KZT = 'Kazakhstani Tenge'
    LAK = 'Laotian Kip'
    LBP = 'Lebanese Pound'
    LKR = 'Sri Lankan Rupee'
    LRD = 'Liberian Dollar'
    LSL = 'Lesotho Loti'
    LTL = 'Lithuanian Litas'
    LVL = 'Latvian Lats'
    LYD = 'Libyan Dinar'
    MAD = 'Moroccan Dirham'
    MDL = 'Moldovan Leu'
    MGA = 'Malagasy Ariary'
    MKD = 'Macedonian Denar'
    MMK = 'Myanmar Kyat'
    MNT = 'Mongolian Tugrik'
    MOP = 'Macanese Pataca'
    MRO = 'Mauritania Ouguiya'
    MUR = 'Mauritian Rupee'
    MVR = 'Maldivian Rufiyaa'
    MWK = 'Malawi Kwacha'
    MXN = 'Mexican Peso'
    MYR = 'Malaysian Ringgit'
    MZN = 'Mozambican Metical'
    NAD = 'Namibian Dollar'
    NGN = 'Nigerian Naira'
    NIO = 'Nicaraguan Córdoba'
    NOK = 'Norwegian Krone'
    NPR = 'Nepalese Rupee'
    NZD = 'New Zealand Dollar'
    OMR = 'Omani Rial'
    PAB = 'Panamanian Balboa'
    PEN = 'Peruvian Nuevo Sol'
    PGK = 'Papua New Guinean Kina'
    PHP = 'Philippine Peso'
    PKR = 'Pakistani Rupee'
    PLN = 'Polish Zloty'
    PYG = 'Paraguayan Guarani'
    QAR = 'Qatari Rial'
    RON = 'Romanian Leu'
    RSD = 'Serbian Dinar'
    RUB = 'Russian Ruble'
    RWF = 'Rwandan Franc'
    SAR = 'Saudi Riyal'
    SBD = 'Solomon Islands Dollar'
    SCR = 'Seychellois Rupee'
    SDG = 'Sudanese Pound'
    SEK = 'Swedish Krona'
    SGD = 'Singapore Dollar'
    SHP = 'Saint Helena Pound'
    SLL = 'Sierra Leonean Leone'
    SOS = 'Somali Shilling'
    SRD = 'Surinamese Dollar'
    STD = 'São Tomé and Príncipe Dobra'
    SVC = 'Salvadoran Colón'
    SYP = 'Syrian Pound'
    SZL = 'Swazi Lilangeni'
    THB = 'Thai Baht'
    TJS = 'Tajikistan Somoni'
    TMT = 'Turkmenistan Manat'
    TND = 'Tunisian Dinar'
    TOP = 'Tongan Paʻanga'
    TRY = 'Turkish Lira'
    TTD = 'Trinidad and Tobago Dollar'
    TWD = 'New Taiwan Dollar'
    TZS = 'Tanzanian Shilling'
    UAH = 'Ukrainian Hryvnia'
    UGX = 'Ugandan Shilling'
    USD = 'United States Dollar'
    UYU = 'Uruguayan Peso'
    UZS = 'Uzbekistan Som'
    VEF = 'Venezuelan Bolívar Fuerte'
    VND = 'Vietnamese Dong'
    VUV = 'Vanuatu Vatu'
    WST = 'Samoan Tala'
    XAF = 'CFA Franc BEAC'
    XAG = 'Silver (troy ounce)'
    XAU = 'Gold (troy ounce)'
    XCD = 'East Caribbean Dollar'
    XDR = 'Special Drawing Rights'
    XOF = 'CFA Franc BCEAO'
    XPF = 'CFP Franc'
    YER = 'Yemeni Rial'
    ZAR = 'South African Rand'
    ZMK = 'Zambian Kwacha (pre-2013)'
    ZMW = 'Zambian Kwacha'
    ZWL = 'Zimbabwean Dollar'


class OperationStatus(Enum, metaclass=_EnumMeta):
    new = auto()
    pending = auto()
    expired = auto()
    completed = auto()
    mismatch = auto()
    error = auto()
    cancelled = auto()


class OperationType(Enum, metaclass=_EnumMeta):
    cash_in = auto()
    cash_out = auto()
    mass_cash_out = auto()
    invoice = auto()


class PlanName(Enum, metaclass=_EnumMeta):
    economy = auto()
    normal = auto()
    priority = auto()
    custom = auto()
